﻿using System;
using System.Linq;
using System.Windows.Forms;
using KeePassLib;
using KeePass.Forms;
using System.Diagnostics;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using KeePass.Plugins;
using System.Reflection;

namespace ITanMaster
{
    public partial class ITanEntryForm : Form
    {
        private string m_strToken = "";
        private Queue<PwEntry> pes = null;
        public string Token { get { return m_strToken; } }
        private MainForm m_mainForm = null;
        private IPluginHost m_host = null;
        private PwGroup m_pgStorage = null;

        private int m_tokensRequested = 1;
        private ITanMaster.TokenProvider m_TokenProvider = TokenProvider.Sequence;

        public ITanEntryForm()
        {
            InitializeComponent();
        }

        public void InitEx(MainForm mainForm, PwGroup pgStorage, ITanMaster.TokenProvider tp, int tokensRequested)
        {
            m_mainForm = mainForm;
            m_pgStorage = pgStorage;
            m_tokensRequested = tokensRequested;
            m_TokenProvider = tp;
            tsStatusLabel.Text = tp + ": Enter " + tokensRequested + " tokens";
            btnOK.Enabled = false;

            //m_mainForm.KeyDown += new KeyEventHandler(this.On_KeyDown);
            //this.KeyDown += new KeyEventHandler(this.On_KeyDown);
            //this.KeyUp += new KeyEventHandler(On_KeyUp);
            //this.txbIndex.KeyDown += new KeyEventHandler(this.On_KeyDown);
            //this.txbToken.KeyDown += new KeyEventHandler(this.On_KeyDown);
            //this.btnOK.KeyDown += new KeyEventHandler(this.On_KeyDown);
            //this.btnOK.KeyUp += new KeyEventHandler(this.On_KeyUp);

            this.bwOK.RunWorkerCompleted += new System.ComponentModel.RunWorkerCompletedEventHandler(bwOK_RunWorkerCompleted);
            this.Refresh();
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            if (keyData == Keys.Return)
            {
                if (btnOK.Enabled)
                {
                    lblREV.Text += "*";
                    ProcessTokens();
                    btnOK.Select();
                    bwOK.RunWorkerAsync();
                }
                return true;
            }
            else
            {
                return base.ProcessCmdKey(ref msg, keyData);
            }
        }

/*
        private void On_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                MessageBox.Show("suppr Enter DN" + sender.GetType());
                e.SuppressKeyPress = true;
            }
        }
        private void On_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                MessageBox.Show("suppr Enter UP" + sender.GetType());
                e.SuppressKeyPress = true;
            }
        }
*/

        private void OnOK_Click(object sender, EventArgs e)
        {
            ProcessTokens();
        }

        private void ProcessTokens()
        {
            if (pes == null) return;
            foreach (PwEntry pe in pes) CheckAndInvalidate(pe);
        }
        
        private void OnIndex_TextChanged(object sender, EventArgs e)
        {
            TextBox tb = (TextBox)sender;
            PwEntry pe = null;
            pes = new Queue<PwEntry>(m_tokensRequested);
            string[] tokens;
            StringBuilder result = new StringBuilder(32);
            string tmp = "";
            txbToken.Text = "";
            //Debug.WriteLine("Parsing '" + tb.Text + "'");
            switch (m_TokenProvider)
            {
                case TokenProvider.InteractiveBrokers:
                    if (!TryParseSingleNumberIndices(tb.Text, out tokens)) return;
                    foreach (string t in tokens)
                    {
                        int idx = 0;
                        if (!int.TryParse(t, out idx)) return;
                        tmp = GetTokenPart(idx, out pe);
                        if (pe != null)
                            pes.Enqueue(pe);
                        else
                            continue;
                        result.Append(tmp); // + " "); //tb 03012012 - eliminate space between 2 IB tokens for Web access
                        txbToken.Text += String.Format("{0}:{1} ", idx, tmp);
                        //Debug.Assert(pe != null, "PwEntry1 is null ??");
                    }
                    break;
                case TokenProvider.Flatex:
                    if (!TryParseFlatexIndices(tb.Text, out tokens)) return;
                    foreach (string t in tokens)
                    {
                        int idx = 0;
                        if (!int.TryParse(t, out idx)) return;
                        tmp = GetTokenPart(idx, out pe);
                        if (pe != null)
                            pes.Enqueue(pe);
                        else
                            continue;
                        result.Append(tmp);
                        txbToken.Text += String.Format("{0}:{1} ", idx, tmp);
                        //Debug.Assert(pe != null, "PwEntry1 is null ??");
                    }
                    break;
                case TokenProvider.FlatexCH:
                    if (!TryParseFlatexCHIndices(tb.Text, out tokens)) return;
                    foreach (string t in tokens)
                    {
                        int idx = 0;
                        if (!int.TryParse(t, out idx)) return;
                        tmp = GetTokenPart(idx, out pe);
                        if (pe != null)
                            pes.Enqueue(pe);
                        else
                            continue;
                        result.Append(tmp);
                        txbToken.Text += String.Format("{0}:{1} ", idx, tmp);
                        //Debug.Assert(pe != null, "PwEntry1 is null ??");
                    }
                    break;

                case TokenProvider.Sequence:
                case TokenProvider.Card:
                default:
                    break;
            }

            m_strToken = result.ToString().Trim();
        }

        private bool TryParseSingleNumberIndices(string strInput, out string[] tokens)
        {
            tokens = strInput.Split(new string[] { " ", @"," }, StringSplitOptions.RemoveEmptyEntries);
            if (tokens.Length <= m_tokensRequested && tokens.Length > 0) return true;
            txbToken.Text = "?";
            return false;
        }

        private bool TryParseFlatexIndices(string strInput, out string[] tokens)
        {
            //recondition the input string - remove whitespace, group in 2char strings
            char[] worker = strInput.Replace(" ","").ToCharArray();
            StringBuilder sb = new StringBuilder();
            int pos = 0;
            foreach (char c in worker)
            {
                sb.Append(c.ToString());
                if (++pos % 2 == 0) sb.Append(" ");
            }
            strInput = sb.ToString().Trim();

            string[] inputTokens = strInput.Split(new string[] { " ", @"," }, StringSplitOptions.RemoveEmptyEntries);
            tokens = null; // new string[] { }; //empty array
            Queue results = new Queue(m_tokensRequested);
            if (inputTokens.Length <= m_tokensRequested && inputTokens.Length > 0)
            {
                foreach (string inp in inputTokens)
                {
                    try
                    {
                        char[] inc = inp.ToCharArray();
                        if (inc.Length != 2) continue;
                        string strRow = inc[0].ToString().ToUpper();
                        if (!Enum.GetNames(typeof(FlatexLines)).Contains(strRow)) throw new ArgumentException();
                        int row = (int)Enum.Parse(typeof(FlatexLines), inc[0].ToString(), true);
                        int col = int.Parse(inc[1].ToString());
                        string idx = (col + 9 * row).ToString();
                        results.Enqueue(idx);
                    }
                    catch
                    {
                        //results.Enqueue("");
                    }
                }
            }
            if (results.Count > 0)
            {
                tokens = new string[results.Count];
                int i = 0;
                foreach (string r in results) tokens[i++] = r;
                return true;
            }
            txbToken.Text = "?";
            return false;
        }

        private bool TryParseFlatexCHIndices(string strInput, out string[] tokens)
        {
            //recondition the input string - remove whitespace
            strInput = strInput.ToUpper();
            char[] worker = strInput.Replace(" ", "").ToCharArray();
            StringBuilder sb = new StringBuilder();
            //put a space before each header character
            foreach (char c in worker)
            {
                if (Enum.GetNames(typeof(FlatexCHHeader)).Contains(c.ToString())) sb.Append(" ");
                sb.Append(c.ToString());
            }
            strInput = sb.ToString().Trim();

            string[] inputTokens = strInput.Split(new string[] { " ", @"," }, StringSplitOptions.RemoveEmptyEntries);
            tokens = null; // new string[] { }; //empty array
            Queue results = new Queue(m_tokensRequested);
            if (inputTokens.Length <= m_tokensRequested && inputTokens.Length > 0)
            {
                foreach (string inp in inputTokens)
                {
                    try
                    {
                        char[] inc = inp.ToCharArray();
                        if (inc.Length > 3) continue;
                        //check A..F for column
                        if (!Enum.GetNames(typeof(FlatexCHHeader)).Contains(inc[0].ToString())) continue;
                        int col = (int)Enum.Parse(typeof(FlatexCHHeader), inc[0].ToString(), true) + 1;
                        //check row for 1..10
                        int row = int.Parse(inp.Substring(1)) - 1;
                        if (row < 0 || row > 9) continue;
                        string idx = (col + 8 * row).ToString();
                        results.Enqueue(idx);
                    }
                    catch
                    {
                        //results.Enqueue("");
                    }
                }
            }
            if (results.Count > 0)
            {
                tokens = new string[results.Count];
                int i = 0;
                foreach (string r in results) tokens[i++] = r;
                return true;
            }
            txbToken.Text = "?";
            return false;
        }

        private string GetTokenPart(int index, out PwEntry pe)
        {
            pe = null;
            if (index < 0) return "?";
            var groupEntries = from e in m_pgStorage.GetEntries(false)
                               where e.Strings.Get(PwDefs.UserNameField).ReadString() == index.ToString()
                               select e;
            if (groupEntries.Count() != 1) return "?";
            pe = groupEntries.ElementAt(0);
            if (IsExpired(pe)) return "expired";
            if (pe == null) return "?";
            return pe.Strings.Get(PwDefs.PasswordField).ReadString();
        }

        private void CheckAndInvalidate(PwEntry pe)
        {
            if (PwDefs.IsTanEntry(pe) == false) return;

            //tb 06012012 KeePass.Util.EntryUtil.ExpireTanEntry(pe);
            PwDatabase pd = m_mainForm.DocumentManager.FindContainerOf(pe);
            KeePass.Util.EntryUtil.ExpireTanEntryIfOption(pe, pd);
            
            m_mainForm.RefreshEntriesList();
            m_mainForm.UpdateUI(false, null, false, null, true, null, true);
        }

        private bool IsExpired(PwEntry pe)
        {
            return (pe.Expires && (pe.ExpiryTime <= DateTime.Now));
        }

        private void txbToken_TextChanged(object sender, EventArgs e)
        {
            TextBox tb = (TextBox)sender;
            btnOK.Enabled = false;
            if (tb.Text.Contains("expired")) return;
            if (tb.Text.Contains("?")) return;
            if (pes == null || m_tokensRequested != pes.Count) return;
            btnOK.Enabled = true;
        }

        private void ITanEntryForm_Load(object sender, EventArgs e)
        {
            var versionAttribute = Assembly.GetExecutingAssembly().GetCustomAttributes(typeof(AssemblyFileVersionAttribute), true).FirstOrDefault() as AssemblyFileVersionAttribute;
            lblREV.Text = versionAttribute.Version;
        }

        void ITanEntryForm_FormClosing(object sender, System.Windows.Forms.FormClosingEventArgs e)
        {
            //m_mainForm.KeyDown -= this.On_KeyDown;
        }

        private void bwOK_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            DateTime trigger = System.DateTime.Now.AddMilliseconds(500); 
            while (System.DateTime.Now < trigger) Application.DoEvents();
        }

        void bwOK_RunWorkerCompleted(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e)
        {
            btnOK.PerformClick();
        }


    }
}
