﻿using System;
using System.Linq;
using KeePass.Plugins;
using System.Diagnostics;
using System.Windows.Forms;
using System.Drawing;
using KeePass.Forms;
using KeePassLib;
using KeePass.UI;
using KeePass.Util;
using System.Text;

namespace ITanMaster
{
    public sealed class ITanMasterDefs
    {
        public const string ITanFirstEntry = "<ITAN1>";
        public const string ITanSecondEntry = "<ITAN2>";
        public const string ITanTitleEntries = "<ITAN>";
    }

    public enum TokenProvider { Sequence, Card, InteractiveBrokers, Flatex, FlatexCH };

    public enum FlatexLines { A, B, C, D, E, F, G, H, K, L, M };

    public enum FlatexCHHeader { A, B, C, D, E, F, G, H };

    public sealed class ITanMasterExt : Plugin
    {
        private IPluginHost m_host = null;

        private MainForm m_mainForm = null;
        private DocumentManagerEx m_docMgr = null;
        private MenuStrip m_mainMenu = null;
        private ToolStripItem m_mainIBKeySelect = null;
        private ToolStripMenuItem m_menuTools = null;
        private ToolStripMenuItem m_toolsITanWizard;
        private ToolStripSeparator m_toolSep = null;

        private TokenProvider m_tokenProvider = TokenProvider.Sequence;

        public override bool Initialize(IPluginHost host)
        {
            m_host = host;
            InitializeComponent();
            return true;
        }

        private void InitializeComponent()
        {
            m_mainForm = m_host.MainWindow;
            m_mainMenu = m_host.MainWindow.MainMenu;
            m_mainIBKeySelect = new ToolStripMenuItem("Select ITan");
            m_mainIBKeySelect.Click += new EventHandler(OnMainMenuITanKeySelect_Click);
            m_mainMenu.Items.Add(m_mainIBKeySelect);

            m_menuTools = m_host.MainWindow.ToolsMenu;
            m_toolSep = new System.Windows.Forms.ToolStripSeparator();
            m_toolsITanWizard = new System.Windows.Forms.ToolStripMenuItem();
            this.m_toolsITanWizard.Image = Properties.Resources.Dollar_16x16;
            this.m_toolsITanWizard.Name = "m_menuITanWizard";
            this.m_toolsITanWizard.Size = new System.Drawing.Size(195, 22);
            this.m_toolsITanWizard.Text = "ITan Key Wizard";
            this.m_toolsITanWizard.Click += new EventHandler(OnIBKeyWizard_Click);
            m_menuTools.DropDownItems.Add(m_toolSep);
            m_menuTools.DropDownItems.Add(m_toolsITanWizard);
            m_menuTools.DropDownOpening += new EventHandler(OnTools_DropDownOpening);

            AutoType.FilterSendPre += OnAutoType_FilterSendPre;
        }

        void OnAutoType_FilterSendPre(object sender, AutoTypeEventArgs e)
        {
            string filter = e.Entry.Strings.Get(PwDefs.TitleField).ReadString();
            if (!filter.StartsWith(ITanMasterDefs.ITanSecondEntry)) return;

            filter = filter.Replace(ITanMasterDefs.ITanSecondEntry,"");
            int tokens = GetTokenProvider(filter);

            Debug.WriteLine("Trapped sequence=" + e.Sequence);
            ITanEntryForm entry = new ITanEntryForm();
            entry.InitEx(m_mainForm, e.Entry.ParentGroup, m_tokenProvider, tokens);
            entry.StartPosition = FormStartPosition.CenterScreen;
            if (entry.ShowDialog() == DialogResult.OK)
            {
                StringBuilder sb = new StringBuilder();
                bool includestart;
                bool includeend;
                bool includealso = false;
                string inp = e.Sequence;
                if (inp != null && inp.Length > 0)
                {
                    int pos = 0;
                    string p = inp.Substring(0, 1);
                    while (pos < e.Sequence.Length)
                    {
                        p = inp.Substring(pos, 1);
                        includestart = p == "{";
                        includeend = p == "}";
                        if (includestart || includealso || includeend)
                        {
                            sb.Append(p);
                        }
                        includealso = includealso ? !includeend : includestart;
                        ++pos;
                    }
                }
                e.Sequence = entry.Token + sb.ToString();
                Debug.WriteLine("Result sequence=" + e.Sequence);
            }
            else e.Sequence = "";
        }

        void OnMainMenuITanKeySelect_Click(object sender, EventArgs e)
        {
            m_docMgr = m_mainForm.DocumentManager;
            if (m_docMgr == null) { Debug.Assert(false); return; }

            PwDatabase pwDb = m_docMgr.ActiveDatabase;
            if (!pwDb.IsOpen) { Debug.Assert(false); return; }

            PwGroup pgSelected = m_mainForm.GetSelectedGroup();

            if (pgSelected == null)
            {
                MessageBox.Show("Select group with iTan support", "No Group selected", MessageBoxButtons.OK);
                return;
            }

            var itanSecondEntries = from ent in pgSelected.GetEntries(true)
                                    where ent.Strings.Get(PwDefs.TitleField).ReadString().StartsWith(ITanMasterDefs.ITanSecondEntry)
                                    select ent.Strings.Get(PwDefs.TitleField).ReadString();

            if (itanSecondEntries.Count() != 1)
            {
                MessageBox.Show("iTan Key Token not found", "No Step 2 token", MessageBoxButtons.OK);
                return;
            }

            //Step 2 token format is <"ITan2"><#ofTokens|TokenProvider>
            string entry2 = itanSecondEntries.ElementAt(0).Replace(ITanMasterDefs.ITanSecondEntry, "");

            int tokens = 1;
            try
            {
                tokens = GetTokenProvider(entry2);
            }
            catch
            {
                MessageBox.Show("iTan Key Token format error", "Illegal step 2 token", MessageBoxButtons.OK);
                return;
            }


            ITanEntryForm entry = new ITanEntryForm();

            entry.InitEx(m_mainForm, pgSelected, m_tokenProvider, tokens);

            if (entry.ShowDialog() == DialogResult.OK)
            {
                if (string.IsNullOrEmpty(entry.Token)) return;
                if (ClipboardUtil.CopyAndMinimize(entry.Token, true, m_mainForm, null, m_docMgr.ActiveDatabase))
                    m_mainForm.StartClipboardCountdown();
            }
        }

        private int GetTokenProvider(string entry2)
        {
            int tokens;
            m_tokenProvider = TokenProvider.Sequence;
            if (!int.TryParse(entry2, out tokens))
            {
                m_tokenProvider = (TokenProvider)Enum.Parse(typeof(TokenProvider), entry2);
                switch (m_tokenProvider)
                {
                    case TokenProvider.InteractiveBrokers:
                        tokens = 2;
                        break;
                    case TokenProvider.Flatex:
                        tokens = 3;
                        break;
                    case TokenProvider.FlatexCH:
                        tokens = 1;
                        break;
                    case TokenProvider.Sequence:
                    case TokenProvider.Card:
                    default:
                        break;
                }
            }
            return tokens;
        }

        void OnTools_DropDownOpening(object sender, EventArgs e)
        {
            m_toolsITanWizard.Enabled =
                m_mainForm.DocumentManager != null &&
                m_mainForm.DocumentManager.ActiveDatabase.IsOpen;
        }

        private void ClearComponent()
        {
            m_menuTools.DropDownItems.Remove(m_toolSep);
            m_menuTools.DropDownItems.Remove(m_toolsITanWizard);
            m_mainMenu.Items.Remove(m_toolsITanWizard);
            AutoType.FilterSendPre -= OnAutoType_FilterSendPre;
        }

        void OnIBKeyWizard_Click(object sender, EventArgs e)
        {
            m_docMgr = m_mainForm.DocumentManager;
            if (m_docMgr == null) { Debug.Assert(false); return; }

            PwDatabase pwDb = m_docMgr.ActiveDatabase;
            if (!pwDb.IsOpen) { Debug.Assert(false); return; }

            PwGroup pgSelected = m_mainForm.GetSelectedGroup();
            if (pgSelected == null) return;

            ITanWizardForm kwf = new ITanWizardForm();
            kwf.InitEx(m_mainForm, pwDb, pgSelected);

            kwf.ShowDialog();
            m_mainForm.UpdateUI(false, null, true, null, true, null, true);
        }

        public override void Terminate()
        {
            ClearComponent();
        }

        public override Image SmallIcon
        {
            get { return Properties.Resources.Dollar_16x16; }
        }


    }
}
